#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from .authorizer import (
    AccessTokenAuthorizer,
    AuthorizationCodeAuthorizer,
    ClientCredentialsAuthorizer,
)
from .connector import OAuthConnector
from .data import AccessToken, AnyScopes, EvoScopes, OAuthScopes, Scopes
from .exceptions import OAuthError
from .oauth_redirect_handler import OAuthRedirectHandler

__all__ = [
    "AccessToken",
    "AccessTokenAuthorizer",
    "AnyScopes",
    "AuthorizationCodeAuthorizer",
    "ClientCredentialsAuthorizer",
    "EvoScopes",
    "OAuthConnector",
    "OAuthError",
    "OAuthRedirectHandler",
    "OAuthScopes",
    "Scopes",
]
