#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID

from evo.discovery import Hub, Organization

BASE_URL = "http://unittest.localhost/"
ACCESS_TOKEN = "<not-a-real-token>"

HUB = Hub(
    code="hub-code",
    display_name="Hub Name",
    url=BASE_URL,
    services=["service1", "service2"],
)
ORG = Organization(id=UUID(int=1234), display_name="Organization Name", hubs=(HUB,))
WORKSPACE_ID = UUID(int=5678)
